# Copyright 2018 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sendgrid
import six
from python_http_client import exceptions
from sendgrid.helpers.mail import Content, Email, Header, Mail, To

from c7n_mailer.utils import decrypt
from c7n_mailer.utils_email import get_mimetext_message, is_email


class SendGridDelivery(object):

    def __init__(self, config, session, logger):
        self.config = config
        self.logger = logger
        self.session = session
        api_key = decrypt(self.config, self.logger, self.session, 'sendgrid_api_key')
        self.sendgrid_client = sendgrid.SendGridAPIClient(api_key)

    def get_to_addrs_sendgrid_messages_map(self, queue_message):
        # eg: { ('milton@initech.com', 'peter@initech.com'): [resource1, resource2, etc] }
        to_addrs_to_resources_map = self.get_email_to_addrs_to_resources_map(queue_message)

        to_addrs_to_content_map = {}
        for to_addrs, resources in six.iteritems(to_addrs_to_resources_map):
            to_addrs_to_content_map[to_addrs] = get_mimetext_message(
                self.config,
                self.logger,
                queue_message,
                resources,
                list(to_addrs)
            )
        # eg: { ('milton@initech.com', 'peter@initech.com'): message }
        return to_addrs_to_content_map

    # this function returns a dictionary with a tuple of emails as the key
    # and the list of resources as the value. This helps ensure minimal emails
    # are sent, while only ever sending emails to the respective parties.
    def get_email_to_addrs_to_resources_map(self, queue_message):
        email_to_addrs_to_resources_map = {}
        targets = queue_message['action']['to']

        for resource in queue_message['resources']:
            # this is the list of emails that will be sent for this resource
            resource_emails = []

            for target in targets:
                if target.startswith('tag:') and 'tags' in resource:
                    tag_name = target.split(':', 1)[1]
                    result = resource.get('tags', {}).get(tag_name, None)
                    if is_email(result):
                        resource_emails.append(result)
                elif is_email(target):
                    resource_emails.append(target)

            resource_emails = tuple(sorted(set(resource_emails)))

            if resource_emails:
                email_to_addrs_to_resources_map.setdefault(resource_emails, []).append(resource)

        if email_to_addrs_to_resources_map == {}:
            self.logger.debug('Found no email addresses, sending no emails.')
        # eg: { ('milton@initech.com', 'peter@initech.com'): [resource1, resource2, etc] }
        return email_to_addrs_to_resources_map

    def sendgrid_handler(self, queue_message, to_addrs_to_email_messages_map):
        self.logger.info("Sending account:%s policy:%s %s:%s email:%s to %s" % (
            queue_message.get('account', ''),
            queue_message['policy']['name'],
            queue_message['policy']['resource'],
            str(len(queue_message['resources'])),
            queue_message['action'].get('template', 'default'),
            to_addrs_to_email_messages_map))

        for email_to_addrs, message in six.iteritems(to_addrs_to_email_messages_map):
            for to_address in email_to_addrs:
                try:
                    mail = SendGridDelivery._sendgrid_mail_from_email_message(message)
                    self.sendgrid_client.send(mail)
                except (exceptions.UnauthorizedError, exceptions.BadRequestsError) as e:
                    self.logger.warning(
                        "\n**Error \nPolicy:%s \nAccount:%s \nSending to:%s \n\nRequest body:"
                        "\n%s\n\nRequest headers:\n%s\n\n mailer.yml: %s" % (
                            queue_message['policy'],
                            queue_message.get('account', ''),
                            email_to_addrs,
                            e.body,
                            e.headers,
                            self.config
                        )
                    )
                    return False
        return True

    @staticmethod
    def _sendgrid_mail_from_email_message(message):
        """
        Create a Mail object from an instance of email.message.EmailMessage.

        This is a copy and tweak of Mail.from_EmailMessage from the SendGrid SDK
        to get around a bug where it creates an Email object and later requires
        that object to be an instance of a To, Cc, or Bcc object.

        It also strips out any reserved key headers on the message object.
        """

        mail = Mail(
            from_email=Email(message.get('From')),
            subject=message.get('Subject'),

            # Create a To object instead of an Email object
            to_emails=To(message.get('To')),
        )
        try:
            body = message.get_content()
        except AttributeError:
            # Python2
            body = message.get_payload(decode=True).decode('utf-8')
        mail.add_content(Content(
            message.get_content_type(),
            body.strip()
        ))

        # These headers are not allowed on the message object
        # https://sendgrid.com/docs/API_Reference/Web_API_v3/Mail/errors.html#message.headers
        skip_headers = [
            'x-sg-id', 'x-sg-eid', 'received', 'dkim-signature', 'Content-Type',
            'Content-Transfer-Encoding', 'To', 'From', 'Subject', 'Reply-To', 'CC', 'BCC'
        ]

        for k, v in message.items():
            if k not in skip_headers:
                mail.add_header(Header(k, v))
        return mail



# Copyright 2019 Microsoft Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from c7n_azure.constants import RESOURCE_VAULT
from azure.keyvault import KeyVaultId


def azure_decrypt(config, logger, session, encrypted_field):
    data = config[encrypted_field]  # type: str
    if type(data) is dict:
        kv_session = session.get_session_for_resource(resource=RESOURCE_VAULT)
        secret_id = KeyVaultId.parse_secret_id(data['secret'])
        kv_client = kv_session.client('azure.keyvault.KeyVaultClient')
        return kv_client.get_secret(secret_id.vault, secret_id.name, secret_id.version).value

    return data



import logging
from email.mime.text import MIMEText
from email.utils import parseaddr

from .utils import (
    get_message_subject, get_rendered_jinja)

logger = logging.getLogger('c7n_mailer.utils.email')


# Those headers are defined as follows:
#  'X-Priority': 1 (Highest), 2 (High), 3 (Normal), 4 (Low), 5 (Lowest)
#              Non-standard, cf https://people.dsv.su.se/~jpalme/ietf/ietf-mail-attributes.html
#              Set by Thunderbird
#  'X-MSMail-Priority': High, Normal, Low
#              Cf Microsoft https://msdn.microsoft.com/en-us/library/gg671973(v=exchg.80).aspx
#              Note: May increase SPAM level on Spamassassin:
#                    https://wiki.apache.org/spamassassin/Rules/MISSING_MIMEOLE
#  'Priority': "normal" / "non-urgent" / "urgent"
#              Cf https://tools.ietf.org/html/rfc2156#section-5.3.6
#  'Importance': "low" / "normal" / "high"
#              Cf https://tools.ietf.org/html/rfc2156#section-5.3.4
PRIORITIES = {
    '1': {
        'X-Priority': '1 (Highest)',
        'X-MSMail-Priority': 'High',
        'Priority': 'urgent',
        'Importance': 'high',
    },
    '2': {
        'X-Priority': '2 (High)',
        'X-MSMail-Priority': 'High',
        'Priority': 'urgent',
        'Importance': 'high',
    },
    '3': {
        'X-Priority': '3 (Normal)',
        'X-MSMail-Priority': 'Normal',
        'Priority': 'normal',
        'Importance': 'normal',
    },
    '4': {
        'X-Priority': '4 (Low)',
        'X-MSMail-Priority': 'Low',
        'Priority': 'non-urgent',
        'Importance': 'low',
    },
    '5': {
        'X-Priority': '5 (Lowest)',
        'X-MSMail-Priority': 'Low',
        'Priority': 'non-urgent',
        'Importance': 'low',
    }
}


def is_email(target):
    if target is None:
        return False
    if target.startswith('slack://'):
        logger.debug("Slack payload, not an email.")
        return False
    if parseaddr(target)[1] and '@' in target and '.' in target:
        return True
    else:
        return False


def priority_header_is_valid(priority_header, logger):
    try:
        priority_header_int = int(priority_header)
    except ValueError:
        return False
    if priority_header_int and 0 < int(priority_header_int) < 6:
        return True
    else:
        logger.warning('mailer priority_header is not a valid string from 1 to 5')
        return False


def set_mimetext_headers(message, subject, from_addr, to_addrs, cc_addrs, priority, logger):
    """Sets headers on Mimetext message"""

    message['Subject'] = subject
    message['From'] = from_addr
    message['To'] = ', '.join(to_addrs)
    if cc_addrs:
        message['Cc'] = ', '.join(cc_addrs)

    if priority and priority_header_is_valid(priority, logger):
        priority = PRIORITIES[str(priority)].copy()
        for key in priority:
            message[key] = priority[key]

    return message


def get_mimetext_message(config, logger, message, resources, to_addrs):
    body = get_rendered_jinja(
        to_addrs, message, resources, logger,
        'template', 'default', config['templates_folders'])

    email_format = message['action'].get('template_format', None)
    if not email_format:
        email_format = message['action'].get(
            'template', 'default').endswith('html') and 'html' or 'plain'

    return set_mimetext_headers(
        message=MIMEText(body, email_format, 'utf-8'),
        subject=get_message_subject(message),
        from_addr=message['action'].get('from', config['from_address']),
        to_addrs=to_addrs,
        cc_addrs=message['action'].get('cc', []),
        priority=message['action'].get('priority_header', None),
        logger=logger
    )
